#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# +--------------------------------------------------------------------------+
# | Layrageu                                                                 |
# | Copyright (C) 2022 Étienne Nadji                                         |
# |                                                                          |
# | This program is free software: you can redistribute it and/or modify     |
# | it under the terms of the GNU Affero General Public License as           |
# | published by the Free Software Foundation, either version 3 of the       |
# | License, or (at your option) any later version.                          |
# |                                                                          |
# | This program is distributed in the hope that it will be useful,          |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of           |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            |
# | GNU Affero General Public License for more details.                      |
# |                                                                          |
# | You should have received a copy of the GNU Affero General Public License |
# | along with this program.  If not, see <https://www.gnu.org/licenses/>.   |
# +--------------------------------------------------------------------------+

"""
Layrageu : send Mastodon toots from a Flat OpenDocuments file.
"""

import os

__version__ = "0.2.1"
__author__ = "Etienne Nadji <etnadji@eml.cc>"

TOOT_VISIBILITIES = ["private", "unlisted", "public"]


def debug_mode() -> True:
    debug = os.environ.get("DEBUG_LAYRAGEU")

    if debug is None:
        return False

    return debug == "1"


# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# +--------------------------------------------------------------------------+
# | Layrageu                                                                 |
# | Copyright (C) 2022 Étienne Nadji                                         |
# |                                                                          |
# | This program is free software: you can redistribute it and/or modify     |
# | it under the terms of the GNU Affero General Public License as           |
# | published by the Free Software Foundation, either version 3 of the       |
# | License, or (at your option) any later version.                          |
# |                                                                          |
# | This program is distributed in the hope that it will be useful,          |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of           |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            |
# | GNU Affero General Public License for more details.                      |
# |                                                                          |
# | You should have received a copy of the GNU Affero General Public License |
# | along with this program.  If not, see <https://www.gnu.org/licenses/>.   |
# +--------------------------------------------------------------------------+

"""
Layrageu : module file for sending toots.
"""

# Imports ===============================================================#

# Standard library ------------------------------

from typing import List, Optional

# Third parties ---------------------------------

from rich.progress import Progress

from mastodon import Mastodon

# Layrageu --------------------------------------

from layrageu.loader import Toot
from layrageu.settings import Settings

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Fonctions =============================================================#


def send_toots(
    instance_name: str,
    settings: Settings,
    toots: List[Toot],
    password: str,
    warning: Optional[str] = None,
    sensitive: bool = False,
    terminal: bool = False,
) -> bool:
    """
    Send toots to Mastodon.

    :type instance_name: str
    :param instance_name: Name of the Mastodon instance in Layrageu settings.
    :type settings: layrageu.settings.Settings
    :param settings: Layrageu settings
    :type toots: List[layrageu.loader.Toot]
    :param toots: List of Mastodon toots to send.
    :type password: str
    :param password: User's password for the Mastodon instance.
    :type warning: Optional[str]
    :param warning: Content warning title.
    :type sensitive: bool
    :param sensitive: Content warning title.
    :rtype: bool
    """

    def send_toot(
        toot_idx: int,
        instance: Mastodon,
        toot_visiblity: str,
        init_toot_visiblity: str,
        toot: Toot,
        warning: Optional[str],
        sensitive: bool,
        reply_to: str,
    ) -> str:
        # Select toot visibily -----------------------

        target = toot_visiblity

        if toot_idx == 0:
            target = init_toot_visiblity

        # --------------------------------------------

        status_data = instance.status_post(
            status=str(toot),
            visibility=target,
            spoiler_text=warning,
            in_reply_to_id=reply_to,
            sensitive=sensitive,
        )

        # Parameters for the next toot to send -------

        reply_to = status_data["id"]

        return reply_to

    # Getting application Token ------------------------------------------

    token = settings.folder / f"layrageu_{instance_name}.secret"

    Mastodon.create_app(
        "Layrageu",
        api_base_url=settings.mi_url(instance_name),
        to_file=token,
    )

    instance = Mastodon(
        client_id=token, api_base_url=settings.mi_url(instance_name)
    )

    # Log in the instance ------------------------------------------------

    try:
        instance.log_in(
            settings.mi_user(instance_name), password, to_file=token
        )
    except Mastodon.MastodonIllegalArgumentError:
        print("Unable to log in Mastodon instance.")
        print("You should try again later.")

        return False

    # Toot visibility ------------------------------------------------

    toot_visiblity = settings.mi_tootv(instance_name)
    init_toot_visiblity = settings.mi_init_tootv(instance_name)

    # Unlisted visibility is public, but initial toot is private.
    # If the user selected private for initial toot, there is no point
    # making the following toots public, even if they are unlisted.
    if init_toot_visiblity == "private":
        toot_visiblity = "private"

    # The first toot ID to reply is None, because each toot responds to
    # the previous one.
    reply_to = None

    # Instance toots must be marked as sensitive ---------------------

    always_sensitive = settings.mi_always_sensitive(instance_name)

    if always_sensitive:
        sensitive = True

    # Sending toots --------------------------------------------------

    if terminal:
        # perc_update = 100 / len(toots)

        with Progress() as progress:
            print()

            sending = progress.add_task("Sending toots…", total=len(toots))

            for idx, toot in enumerate(toots):
                reply_to = send_toot(
                    idx,
                    instance,
                    toot_visiblity,
                    init_toot_visiblity,
                    toot,
                    warning,
                    sensitive,
                    reply_to,
                )

                progress.update(sending, advance=1)

        return True

    for idx, toot in enumerate(toots):
        reply_to = send_toot(
            idx,
            instance,
            toot_visiblity,
            init_toot_visiblity,
            toot,
            warning,
            sensitive,
            reply_to,
        )

    return True


# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# +--------------------------------------------------------------------------+
# | Layrageu                                                                 |
# | Copyright (C) 2022 Étienne Nadji                                         |
# |                                                                          |
# | This program is free software: you can redistribute it and/or modify     |
# | it under the terms of the GNU Affero General Public License as           |
# | published by the Free Software Foundation, either version 3 of the       |
# | License, or (at your option) any later version.                          |
# |                                                                          |
# | This program is distributed in the hope that it will be useful,          |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of           |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            |
# | GNU Affero General Public License for more details.                      |
# |                                                                          |
# | You should have received a copy of the GNU Affero General Public License |
# | along with this program.  If not, see <https://www.gnu.org/licenses/>.   |
# +--------------------------------------------------------------------------+

"""
Layrageu : FODT → Mastodon toots converter.
"""

# Imports ===============================================================#

# Standard library ------------------------------

from __future__ import annotations

import textwrap

from pathlib import Path
from typing import List, Union, Optional

# Third parties ---------------------------------

import lxml.etree as ET

from rich.panel import Panel
from rich.console import Console
from rich.markdown import Markdown

# Layrageu --------------------------------------

from layrageu.fodt2text import FODT_TO_TEXT_XML

# Variables globales 2 / 2 ==============================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

StringOrPath = Union[str, Path]

# Classes ===============================================================#


class Toot:
    """
    Mastodon toot.

    Handles paragraphs and maximum length.
    """

    def __init__(self, paras: Optional[List[Paragraph]] = None):
        self.paras = []

        if paras is not None:
            for para in paras:
                self.paras.append(para)

    def as_markdown(self) -> Markdown:
        """
        Return toot as rich Markdown object.
        """

        paras_md = [para.as_markdown() for para in self.paras]
        source = "\n\n".join(paras_md)

        return Markdown(source)

    def as_html(self) -> str:
        """
        Return toot as HTML string (<p> elements).
        """

        paras_md = [para.as_html() for para in self.paras]
        source = "\n".join(paras_md)

        return source

    def __eq__(self, other):
        if len(self.paras) != len(other.paras):
            return False

        for idx, own_para in enumerate(self.paras):
            if str(own_para) != str(other.paras[idx]):
                return False

        return True

    def add_para(
        self, para: Paragraph, toot_length: int = 500
    ) -> Union[bool, List[Toot]]:
        """
        Add paragraph. Returns a list of toots if the maximum tooth length
        is execeded.

        :type para: Paragraph
        :type toot_length: int
        """

        def create_next_toots(ref_toot, toot_length=500):
            nexts = ref_toot.add_para(para, toot_length)
            return [ref_toot] + nexts

        following_toots = []

        current_length = len(self)

        # This toot length is already maximum toot length. Don't add the
        # paragraph and return the following new toots
        if current_length == toot_length:
            next_toots = create_next_toots(self, toot_length)
            return next_toots

        if current_length < toot_length:
            # Get the length of this toot if we add the full paragraph to
            # it.
            tmp_toot = Toot([para])
            with_full_next_para = current_length + len(tmp_toot)

            # If toot + paragraph >= maximum toot_length
            # => add the paragraph, returns nothing
            if with_full_next_para <= toot_length:
                self.paras.append(para)
                return False

            remaining_length = toot_length - current_length

            if len(para.lines) > 1:
                # The paragraph is too long and has multiple lines ----------

                new_paragraph = Paragraph()
                nexts_line_index = 0

                for idx, line in enumerate(para.lines):

                    remaining_length -= len(new_paragraph)

                    if len(line) <= remaining_length - 2:
                        new_paragraph.lines.append(line)
                        nexts_line_index = idx + 1
                        break

                self.paras.append(new_paragraph)

                next_lines = para.lines[nexts_line_index:]

                if next_lines:
                    new_toot = Toot()
                    merge_paragraph = Paragraph(next_lines)

                    nexts = new_toot.add_para(merge_paragraph, toot_length)

                    following_toots.append(new_toot)

                    if nexts:
                        following_toots += nexts

                    return following_toots

                return False

            # The paragraph is too long and has only one line -----------

            # Split the line nicely

            splited_line = textwrap.wrap(
                # para.lines[0], remaining_length - len("\n\n")
                para.lines[0],
                remaining_length - 2,
                replace_whitespace=False,
            )

            # Add the part of the line that can be merged without
            # exceding maximum toot length
            merge_paragraph = Paragraph([splited_line[0]])
            self.paras.append(merge_paragraph)

            # Create a new paragraph with the remaining part of the line,
            # add it to a new toot and add it to the following toots
            # returned.
            new_paragraph = Paragraph([" ".join(splited_line[1:])])

            new_toot = Toot()
            nexts = new_toot.add_para(new_paragraph, toot_length)

            following_toots.append(new_toot)

            if not nexts:
                return following_toots

            following_toots += nexts

        return following_toots

    def __len__(self) -> int:
        return len(str(self))

    def __str__(self) -> str:
        return "\n\n".join([str(para) for para in self.paras])


class Paragraph:
    """
    Paragraph in a Mastodon toot.
    """

    def __init__(self, lines: Optional[List[str]] = None):
        self.lines = []

        if lines is not None:
            self.lines = lines

    def __repr__(self) -> str:
        return f"{len(self.lines)}|" + "{lb}".join(self.lines)

    def __len__(self) -> int:
        return len(str(self))

    def __str__(self) -> str:
        return "\n".join(self.lines)

    def as_markdown(self) -> str:
        """
        Returns paragraph as a Markdown string.
        Handles line breaks.
        """
        return "\\\n".join(self.lines)

    def as_html(self) -> str:
        """
        Returns paragraph as an HTML paragraph string.
        """
        lines = "<br/>".join(self.lines)
        return f"<p>{lines}</p>"


# Fonctions =============================================================#


def thread_to_toots(
    raw_thread: List[List[str]],
    toot_length=500,
) -> List[Toot]:
    """
    Converts a raw thread into a sequence of Toots.

    :type raw_thread: List[List[str]]
    :type toot_length: int
    :rtype: List[Toot]
    """

    def is_raw_toot_break(data):
        return data == ["--"]

    sequence = []

    for thread_item in raw_thread:

        if not sequence:

            if is_raw_toot_break(thread_item):
                continue

            toot = Toot()
            para = Paragraph(thread_item)

            following_toots = toot.add_para(para, toot_length)

            sequence.append(toot)

            if following_toots:
                for next_toot in following_toots:
                    sequence.append(next_toot)

            continue

        if is_raw_toot_break(thread_item):
            toot = Toot()
            sequence.append(toot)

            continue

        para = Paragraph(thread_item)
        following_toots = sequence[-1].add_para(para, toot_length)

        if following_toots:
            for next_toot in following_toots:
                sequence.append(next_toot)

    return sequence


def raw_thread(thread_input: str) -> List[List[str]]:
    """
    Converts raw text document into a list of elements.

    - Paragraphs are transformed into a List[str].
    - Toots breaks (`--` paragraph) become ["--"]

    :rtype: List[List[str]]
    """
    # First, spliting paragraphs

    thread_input = thread_input.split("\n")

    paras = []

    for text in thread_input:
        # No line break in that paragraph

        if "{lb}" not in text:
            paras.append([text])
            continue

        # Line break found in that paragraph

        text = text.replace("{lb}", "\n")
        paras += [[part for part in text.split("\n") if part != [""]]]

    # At this point the thread is a List[List[str]], strings are joined with
    # line breaks.

    return [part for part in paras if part != [""]]


def load_fodt_converter() -> ET.XSLT:
    """
    Returns Layrageu FODT ⇒ text XSLT converter.
    """

    # stylesheet_path = Path(__file__).resolve(True).parent
    # stylesheet_path = stylesheet_path / "fodt2text.xsl"

    return ET.XSLT(ET.fromstring(FODT_TO_TEXT_XML))


def convert(document_path: StringOrPath, debug: bool = False) -> List[Toot]:
    if debug:
        console = Console()

    # Get document path ----------------------------------------

    if isinstance(document_path, str):
        document_path = Path(document_path)

    document_path = str(document_path.resolve(True))

    # Parse and convert document -------------------------------

    document = ET.parse(document_path)
    converter = load_fodt_converter()
    thread_input = str(converter(document)).rstrip()

    if debug:
        console.print(Panel(thread_input, title="XSLT output"))

    # Transforms document into individual toots ----------------

    splitted = raw_thread(thread_input)

    sequence = unduplicate(thread_to_toots(splitted))

    if debug:
        console.print(
            Panel(
                "\n\n".join(
                    [
                        f"SEQUENCE ITEM {idx} | {len(element)}\n{element}"
                        for idx, element in enumerate(sequence)
                    ]
                ),
                title="Toot sequence",
            )
        )

    # ----------------------------------------------------------

    return sequence


def unduplicate(sequence: List[Toot]) -> List[Toot]:
    """
    Remove possible duplicates in `sequence`.
    """

    new_sequence = []

    for item in sequence:
        add = True

        for old_item in new_sequence:
            if item == old_item:
                add = False
                break

        if add:
            new_sequence.append(item)

    return new_sequence


# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# +--------------------------------------------------------------------------+
# | Layrageu                                                                 |
# | Copyright (C) 2022 Étienne Nadji                                         |
# |                                                                          |
# | This program is free software: you can redistribute it and/or modify     |
# | it under the terms of the GNU Affero General Public License as           |
# | published by the Free Software Foundation, either version 3 of the       |
# | License, or (at your option) any later version.                          |
# |                                                                          |
# | This program is distributed in the hope that it will be useful,          |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of           |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            |
# | GNU Affero General Public License for more details.                      |
# |                                                                          |
# | You should have received a copy of the GNU Affero General Public License |
# | along with this program.  If not, see <https://www.gnu.org/licenses/>.   |
# +--------------------------------------------------------------------------+

"""
Layrageu : send Mastodon toots from Flat OpenDocument files. GUI.
"""

# Imports ===============================================================#

# Standard library ------------------------------

import os
import sys
import copy
import tempfile
import webbrowser

from functools import partial
from pathlib import Path
from typing import Optional, NoReturn

# Third parties ---------------------------------------------------

from appdirs import user_config_dir

# Tkinter -------------------------------------------------

import tkinter

from tkinter import ttk as ttk
from tkinter import filedialog as tkdialog
from tkinter import messagebox as messagebox

# ttkSimpleDialog -----------------------------------------

from ttkSimpleDialog import ttkSimpleDialog

# Layrageu --------------------------------------------------------

from layrageu import TOOT_VISIBILITIES, debug_mode
from layrageu.loader import convert
from layrageu.sender import send_toots
from layrageu.settings import Settings, slugify_instance_url

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"
__version__ = "0.2.1"

PREVIEW_DOC = """
<!DOCTYPE html>
<html>
    <head>
        <meta charset=utf-8 />
        <title>Toot thread preview</title>
        <style>
            html { background-color: #191b22; color: white; }
            body {
                margin: 1em 0;
                display: grid;
                grid-template-columns: 1fr 4fr 1fr;
            }
            main { grid-column: 2; }
            @media screen and (max-width: 800px) {
                body { grid-template-columns: 1em 4fr 1em; }
            }
            article {
              font-size: 1em;
              line-height: 1.5;
              font-family: sans-serif;
              text-align: justify;
              margin-bottom: 1em;
              background-color: #282c37;
              padding: 5px 10px;
              border-radius: 5px;
            }
            article h1 {
                font-size: medium;
                text-align: center;
                margin: 0;
                font-weight: bold;
            }
        </style>
    </head>
    <body>
        <main>
        {TOOTS}
        </main>
    </body>
</html>
"""

TOOT_PREVIEW = """\n<article>
<h1>Toot #{TOOT_IDX} ({TOOT_LENGTH})</h1><section>{TOOT_CONTENT}</section>
</article>\n"""

# Functions =============================================================#


def validate_missing(
    url: str, usermail: str, init_tootv: str, other_tootv: str
):
    """
    Common function to check Mastodon instance configuration form validity.
    """

    missing, message = False, "Some infos are missing:\n"

    for data, msg in [
        [url, "  – Instance url"],
        [usermail, "  – User mail"],
    ]:
        if not len(data):
            missing = True
            message += msg

    if missing:
        return False, message

    return True, ""


# Classes ===============================================================#


class NewInstanceWindow(tkinter.Toplevel):
    """
    New instance window.
    """

    def __init__(self, settings: Settings, app_frame: ttk.Frame):
        tkinter.Toplevel.__init__(self)

        self.app = app_frame
        self.settings = settings

        self.init_ui()

    def init_ui(self) -> NoReturn:
        """
        Initialize window's UI.
        """

        # Instance URL ---------------------------------------------------

        url_label = ttk.Label(self, text="Instance URL")
        self.url_entry = ttk.Entry(self)

        # Instance user mail ---------------------------------------------

        user_label = ttk.Label(self, text="User (mail)")
        self.user_entry = ttk.Entry(self)

        # Toots visibilities ---------------------------------------------

        # Initial toot -----------------------------------------

        initv_label = ttk.Label(self, text="Initial toot visibility")
        self.initv_combo = ttk.Combobox(self, values=TOOT_VISIBILITIES)
        self.initv_combo.current(TOOT_VISIBILITIES.index("public"))

        # Following toots --------------------------------------

        otherv_label = ttk.Label(self, text="Next toots visibility")
        self.otherv_combo = ttk.Combobox(self, values=TOOT_VISIBILITIES)
        self.otherv_combo.current(TOOT_VISIBILITIES.index("unlisted"))

        # Packing everything ---------------------------------------------

        ui_row = 2

        for label, widget in [
            [url_label, self.url_entry],
            [user_label, self.user_entry],
            [initv_label, self.initv_combo],
            [otherv_label, self.otherv_combo],
        ]:
            label.grid(column=1, row=ui_row, padx=5, pady=5, sticky=tkinter.E)
            widget.grid(
                column=2, row=ui_row, padx=5, columnspan=2, sticky=tkinter.W
            )
            ui_row += 1

        ui_row += 1

        ok_btn = ttk.Button(self, text="OK", command=self.validate)
        ok_btn.grid(column=2, row=ui_row, padx=5, pady=5, sticky=tkinter.E)

        cancel_btn = ttk.Button(self, text="Cancel", command=self.cancel)
        cancel_btn.grid(
            column=1,
            row=ui_row,
            padx=5,
            pady=5,
            columnspan=2,
            sticky=tkinter.W,
        )

    def validate(self) -> NoReturn:
        """
        Callback for OK button.
        """

        validated = False

        url = self.url_entry.get()
        usermail = self.user_entry.get()
        init_tootv = self.initv_combo.get()
        other_tootv = self.otherv_combo.get()

        validated, message = validate_missing(
            url, usermail, init_tootv, other_tootv
        )

        if not validated:
            messagebox.showwarning(title="Missing infos", message=message)
            return

        instance_name = slugify_instance_url(url)

        if self.settings.mi_exists(instance_name):
            messagebox.showwarning(
                title=f"Instance already exists",
                message=f"“{instance_name}” instance already exists.",
            )
            return

        self.settings.add_instance(
            instance_name,
            url,
            usermail,
            init_tootv,
            other_tootv,
        )

        self.settings.save()
        self.app.refresh_settings()

        self.destroy()

    def cancel(self) -> NoReturn:
        """
        Callback for Cancel button.
        """
        self.destroy()


class FirstTimeSettings(tkinter.Toplevel):
    """
    First-time launch settings window.
    """

    def __init__(self, settings: Settings, app_frame: ttk.Frame):
        tkinter.Toplevel.__init__(self)

        self.app = app_frame
        self.settings = settings

        self.init_ui()

    def init_ui(self) -> NoReturn:
        """
        Initialize window's UI.
        """

        intro_label = ttk.Label(
            self,
            text="\n".join(
                [
                    "Please enter the details of your Mastodon",
                    "instance access.",
                    "",
                ]
            ),
        )
        intro_label.grid(
            column=1, row=1, padx=5, pady=5, columnspan=2, sticky=tkinter.EW
        )

        # Instance URL ---------------------------------------------------

        url_label = ttk.Label(self, text="Instance URL")
        self.url_entry = ttk.Entry(self)
        self.url_entry.insert(0, self.settings.mi_url("default"))

        # Instance user mail ---------------------------------------------

        user_label = ttk.Label(self, text="User (mail)")
        self.user_entry = ttk.Entry(self)

        # Toots visibilities ---------------------------------------------

        # Initial toot -----------------------------------------

        initv_label = ttk.Label(self, text="Initial toot visibility")
        self.initv_combo = ttk.Combobox(self, values=TOOT_VISIBILITIES)
        self.initv_combo.current(
            TOOT_VISIBILITIES.index(self.settings.mi_init_tootv("default"))
        )

        # Following toots --------------------------------------

        otherv_label = ttk.Label(self, text="Next toots visibility")
        self.otherv_combo = ttk.Combobox(self, values=TOOT_VISIBILITIES)
        self.otherv_combo.current(
            TOOT_VISIBILITIES.index(self.settings.mi_tootv("default"))
        )

        # Packing everything ---------------------------------------------

        ui_row = 2

        for label, widget in [
            [url_label, self.url_entry],
            [user_label, self.user_entry],
            [initv_label, self.initv_combo],
            [otherv_label, self.otherv_combo],
        ]:
            label.grid(column=1, row=ui_row, padx=5, pady=5, sticky=tkinter.E)
            widget.grid(
                column=2, row=ui_row, padx=5, columnspan=2, sticky=tkinter.W
            )
            ui_row += 1

        ui_row += 1

        ok_btn = ttk.Button(self, text="OK", command=self.validate)
        ok_btn.grid(column=2, row=ui_row, padx=5, pady=5, sticky=tkinter.E)

        cancel_btn = ttk.Button(self, text="Cancel", command=self.cancel)
        cancel_btn.grid(
            column=1,
            row=ui_row,
            padx=5,
            pady=5,
            columnspan=2,
            sticky=tkinter.W,
        )

    def validate(self) -> NoReturn:
        """
        Callback for OK button.
        """

        validated = False

        url = self.url_entry.get()
        usermail = self.user_entry.get()
        init_tootv = self.initv_combo.get()
        other_tootv = self.otherv_combo.get()

        validated, message = validate_missing(
            url, usermail, init_tootv, other_tootv
        )

        if not validated:
            messagebox.showwarning(title="Missing infos", message=message)
            return

        self.settings.instances["default"].url = url
        self.settings.instances["default"].usermail = usermail
        self.settings.instances["default"].first_tootv = init_tootv
        self.settings.instances["default"].other_tootv = other_tootv

        self.settings.save()
        self.app.refresh_settings()

        self.destroy()

    def cancel(self) -> NoReturn:
        """
        Callback for Cancel button.
        """
        self.destroy()
        self.app.quit()


class InstancesWindow(tkinter.Toplevel):
    """
    Settings window for Layrageu GUI.
    """

    def __init__(self, settings: Settings, app_frame: ttk.Frame):
        tkinter.Toplevel.__init__(self)

        self.app = app_frame
        self.settings = settings
        self.instances_ui = {}

        self.init_ui()

    def init_ui(self) -> NoReturn:
        """
        Initialize window's UI.
        """

        self.notebook = ttk.Notebook(self)
        self.notebook.grid(
            column=1, row=1, padx=5, pady=5, columnspan=2, sticky=tkinter.W
        )

        for tab_id, instance in enumerate(self.settings.instances.keys()):
            self.instance_frame(instance, tab_id)

        ok_btn = ttk.Button(self, text="OK", command=self.validate)
        ok_btn.grid(column=2, row=2, padx=5, pady=5, sticky=tkinter.E)

        cancel_btn = ttk.Button(self, text="Cancel", command=self.cancel)
        cancel_btn.grid(
            column=1, row=2, padx=5, pady=5, columnspan=2, sticky=tkinter.W
        )

    def cancel(self) -> NoReturn:
        """
        Callback for Cancel button.
        """
        self.destroy()

    def validate(self) -> NoReturn:
        """
        Callback for OK button.
        """

        close = True

        for instance in self.settings.instances.keys():

            if instance not in self.instances_ui:
                continue

            url = self.instances_ui[instance]["url"].get()
            usermail = self.instances_ui[instance]["user"].get()
            init_tootv = self.instances_ui[instance]["init_tootv"].get()
            other_tootv = self.instances_ui[instance]["tootv"].get()

            validated, message = validate_missing(
                url, usermail, init_tootv, other_tootv
            )

            if not validated:
                messagebox.showwarning(
                    title="Missing infos : {instance} instance",
                    message=message,
                )
                close = False
                continue

            self.settings.instances[instance].url = url
            self.settings.instances[instance].usermail = usermail
            self.settings.instances[instance].first_tootv = init_tootv
            self.settings.instances[instance].other_tootv = other_tootv
            self.settings.save()

        if close:
            self.app.refresh_settings()
            self.destroy()

    def delete_instance(self, instance_name: str) -> NoReturn:
        """
        Tells Layrageu settings to delete instance called `instance_name`.
        """

        result = messagebox.askquestion(
            f"Delete “{instance_name}” instance ?",
            f"Are you sure you want to delete “{instance_name}” instance ?",
        )

        if result == "yes":
            self.notebook.forget(self.instances_ui[instance_name]["tab_id"])
            self.instances_ui.pop(instance_name)

            self.settings.mi_delete(instance_name)

            self.app.refresh_instances()

            if len(self.settings.instances) == 1:
                self.destroy()

    def instance_frame(self, instance_name: str, tab_id: int) -> NoReturn:
        """
        Create an instance settings Frame to put in the Instances window
        notebook.
        """

        instance_frame = ttk.Frame(self.notebook)

        self.instances_ui[instance_name] = {
            "name": instance_name,
            "frame": instance_frame,
            "tab_id": tab_id,
        }

        # Instance URL ---------------------------------------------------

        url_label = ttk.Label(instance_frame, text="Instance URL")
        url_entry = ttk.Entry(instance_frame)
        url_entry.insert(0, self.settings.mi_url(instance_name))
        self.instances_ui[instance_name]["url"] = url_entry

        # Instance user mail ---------------------------------------------

        user_label = ttk.Label(instance_frame, text="User (mail)")
        user_entry = ttk.Entry(instance_frame)
        user_entry.insert(0, self.settings.mi_user(instance_name))
        self.instances_ui[instance_name]["user"] = user_entry

        # Toots visibilities ---------------------------------------------

        # Initial toot -----------------------------------------

        initv_label = ttk.Label(instance_frame, text="Initial toot visibility")
        initv_combo = ttk.Combobox(instance_frame, values=TOOT_VISIBILITIES)
        initv_combo.current(
            TOOT_VISIBILITIES.index(self.settings.mi_init_tootv(instance_name))
        )
        self.instances_ui[instance_name]["init_tootv"] = initv_combo

        # Following toots --------------------------------------

        otherv_label = ttk.Label(instance_frame, text="Next toots visibility")
        otherv_combo = ttk.Combobox(instance_frame, values=TOOT_VISIBILITIES)
        otherv_combo.current(
            TOOT_VISIBILITIES.index(self.settings.mi_tootv(instance_name))
        )
        self.instances_ui[instance_name]["tootv"] = otherv_combo

        # Delete instance button -----------------------------------------

        delete_btn = ttk.Button(
            instance_frame,
            text="Delete instance",
            command=partial(self.delete_instance, instance_name),
        )

        # Packing everything ---------------------------------------------

        ui_row = 1

        for label, widget in [
            [url_label, url_entry],
            [user_label, user_entry],
            [initv_label, initv_combo],
            [otherv_label, otherv_combo],
        ]:
            label.grid(column=1, row=ui_row, padx=5, pady=5, sticky=tkinter.E)
            widget.grid(
                column=2, row=ui_row, padx=5, columnspan=2, sticky=tkinter.W
            )
            ui_row += 1

        delete_btn.grid(column=2, row=ui_row, padx=5, pady=5, sticky=tkinter.W)

        # Add instance tab -----------------------------------------------

        self.notebook.add(instance_frame, text=instance_name)


class Layrageu(ttk.Frame):
    """
    Main window of Layrageu GUI.
    """

    def __init__(
        self,
        master: Optional[tkinter.Tk] = None,
    ):
        ttk.Frame.__init__(self, master)

        self.debug_mode = debug_mode()

        # Settings ---------------------------------------------

        self.settings = Settings(user_config_dir("layrageu"))

        # Window style -----------------------------------------

        style = ttk.Style(self)

        if sys.platform.startswith("linux"):
            style.theme_use("clam")

        # GUI data ---------------------------------------------

        self.document = None
        self.toots = []

        # ------------------------------------------------------

        self.master = master

        self.ui_vars = {"sensitive": tkinter.IntVar()}

        self.init_ui()
        self.init_menu()

        if self.settings.needs_first_user_action():
            wizard_window = FirstTimeSettings(self.settings, self)
            wizard_window.title("First time configuration")

    def init_menu(self) -> NoReturn:
        """
        Initialize menus.
        """

        def new_menu(parent, menu_items):
            new_menu = tkinter.Menu(parent)

            for item in menu_items:
                if item[0] == "<separator>":
                    new_menu.add_separator()
                else:
                    new_menu.add_command(label=item[0], command=item[1])

            return new_menu

        app_menu = tkinter.Menu(self.master)
        self.master.config(menu=app_menu)

        file_menu = new_menu(
            app_menu,
            [
                # ["<separator>", None],
                ["Quit", self.quit],
            ],
        )

        instances_menu = new_menu(
            app_menu,
            [
                ["Configure", self.open_instances],
                ["New", self.new_instance],
            ],
        )

        help_menu = new_menu(
            app_menu,
            [
                ["About", self.about],
            ],
        )

        instance_menu_label = {True: "Instances", False: "Instance"}[
            self.settings.instances_count() > 1
        ]

        for submenu in [
            ["File", file_menu],
            [instance_menu_label, instances_menu],
            ["Help", help_menu],
        ]:
            app_menu.add_cascade(
                label=submenu[0],
                menu=submenu[1],
            )

    def init_ui(self) -> NoReturn:
        """
        Initialize Layrageu's UI.
        """

        # Main frame (we want some padding around everything) ------------

        self.main_frame = ttk.Frame(self.master)
        self.main_frame.grid(padx=5, pady=5)

        # Document -------------------------------------------------------

        doc_btn = ttk.Button(
            self.main_frame, text="Document", command=self.select_document
        )
        doc_btn.grid(column=1, row=1, padx=5)

        self.doc_entry = ttk.Entry(self.main_frame)
        self.doc_entry.grid(
            column=2, row=1, padx=5, columnspan=2, sticky=tkinter.W
        )

        # Content warning ------------------------------------------------

        cw_label = ttk.Label(self.main_frame, text="Warning")
        cw_label.grid(column=1, row=2, padx=5, pady=5, sticky=tkinter.E)

        self.cw_entry = ttk.Entry(self.main_frame)
        self.cw_entry.grid(
            column=2, row=2, padx=5, columnspan=2, sticky=tkinter.W
        )

        # Instances ------------------------------------------------------

        instances_label = ttk.Label(self.main_frame, text="Send on instance")
        instances_label.grid(column=1, row=3, padx=5, pady=5, sticky=tkinter.E)

        self.instances_combo = ttk.Combobox(self.main_frame)
        self.instances_combo.grid(
            column=2, row=3, padx=5, pady=5, columnspan=2, sticky=tkinter.W
        )
        self.refresh_instances()

        # Sensible posts -------------------------------------------------

        self.sensitive_check = ttk.Checkbutton(
            self.main_frame,
            text="Sensible",
            variable=self.ui_vars["sensitive"],
        )
        self.sensitive_check.grid(column=1, row=4, pady=5)

        # Send and preview buttons ---------------------------------------

        send_btn = ttk.Button(
            self.main_frame, text="Send toots", command=self.send
        )
        send_btn.grid(column=2, row=5, columnspan=3, padx=5, sticky=tkinter.EW)

        prev_btn = ttk.Button(
            self.main_frame, text="Preview", command=self.preview
        )
        prev_btn.grid(column=1, row=5, padx=5)

    def open_instances(self) -> NoReturn:
        """
        Open the Settings window.
        """
        instances_window = InstancesWindow(self.settings, self)
        instances_window.title("Instances")

    def new_instance(self) -> NoReturn:
        """
        Open the New instance window.
        """
        instances_window = NewInstanceWindow(self.settings, self)
        instances_window.title("New instance")

    def refresh_instances(self) -> NoReturn:
        self.instances_combo["values"] = self.settings.instances_names()
        self.instances_combo.delete(0, "end")
        self.instances_combo.current(0)

    def refresh_settings(self) -> NoReturn:
        """
        Refresh settings when InstancesWindow window is destroyed with
        `InstancesWindow.validate`.
        """
        self.refresh_instances()

    def about(self) -> NoReturn:
        """
        Layrageu's About window.
        """

        messagebox.showinfo(
            "About Layrageu",
            "\n".join(
                [
                    f"Layrageu {__version__}",
                    "",
                    "Copyright © 2022 Étienne Nadji",
                    "",
                    "Layrageu is published under GNU Affero General Public "
                    "License v3 or later (AGPLv3+)",
                ]
            ),
        )

        return True

    def quit(self) -> NoReturn:
        """
        Quit Layrageu GUI.
        """
        sys.exit()

    def select_document(self) -> NoReturn:
        document = tkdialog.askopenfile().name
        self.document = Path(document).resolve(True)

        self.toots = []

        self.doc_entry.delete(0, "end")
        self.doc_entry.insert(0, self.document.name)

    def load_toots(self) -> bool:
        """
        Load Mastodon thread toots.

        :rtype: bool
        :returns: True if toots were loaded.
        """

        if self.document is None:
            return False

        if self.toots:
            return True

        # Load toots from Flat OpenDocument files
        self.toots = convert(self.document)

        return True

    def preview(self) -> bool:
        """
        Preview loaded document's Mastodon thread.

        :rtype: bool
        """

        # Check if there is toots to use ---------------------------------

        toots_loaded = self.load_toots()

        if not toots_loaded:
            return False

        # Create an HTML preview file ------------------------------------

        preview_doc = copy.deepcopy(PREVIEW_DOC)
        toots_content = ""

        for idx, toot in enumerate(self.toots):
            toot_content = copy.deepcopy(TOOT_PREVIEW)

            toot_content = toot_content.replace("{TOOT_IDX}", str(idx + 1))
            toot_content = toot_content.replace(
                "{TOOT_LENGTH}", str(len(toot))
            )
            toot_content = toot_content.replace(
                "{TOOT_CONTENT}", toot.as_html()
            )

            toots_content += toot_content

        preview_doc = preview_doc.replace("{TOOTS}", toots_content)

        # Write and open the temporary file ------------------------------

        webpreview = tempfile.NamedTemporaryFile(
            mode="w", suffix=".html", delete=False
        )
        webpreview.write(preview_doc)

        url = f"file://{webpreview.name}"

        webpreview.close()

        # ----------------------------------------------------------------

        webbrowser.open_new_tab(url)

        return True

    def send(self) -> bool:
        """
        Send loaded document's Mastodon thread.

        :rtype: bool
        """

        # Check if there is toots to use ---------------------------------

        toots_loaded = self.load_toots()

        if not toots_loaded:
            return False

        # Prepare toots send ---------------------------------------------

        # Get instance name
        instance_name = self.instances_combo.get()

        if not self.settings.mi_exists(instance_name):
            return False

        # Password ---------------------------------------------

        password = ttkSimpleDialog.askstring(
            title="Layrageu",
            prompt="Enter your Mastodon's instance password.",
            show="*",
        )

        if password is None or not password:
            messagebox.showerror(
                title="Empty password",
                message="You entered an empty password.",
            )
            return False

        # Content warning --------------------------------------

        content_warning = self.cw_entry.get().rstrip()

        if not content_warning:
            content_warning = None

        # Send toots -----------------------------------------------------

        result = send_toots(
            instance_name,
            self.settings,
            self.toots,
            password,
            content_warning,
            bool(self.ui_vars["sensitive"].get()),
        )

        # Open Mastodon instance in case of sending success --------------

        if result:
            webbrowser.open_new_tab(self.settings.mi_url(instance_name))

        return True


# Fonctions =============================================================#


def starting_dimensions() -> str:
    os_window_dims = {
        "win": [345, 175],
        "linux": [345, 175],
        "darwin": [345, 175],
    }

    # MacOS always needs bigger dimensions, so this a good fallback
    dimensions = os_window_dims["darwin"]

    try:
        dimensions = os_window_dims[sys.platform]
    except IndexError:
        for platform, sizes in os_window_dims.items():
            if not sys.platform.startswith(platform):
                continue

            dimensions = sizes

    return f"{dimensions[0]}x{dimensions[1]}"


def main() -> NoReturn:
    """
    Layrageu GUI entry point.
    """

    # --- Tkinter / Layrageu startup -------------------------------------

    env = tkinter.Tk()
    application = Layrageu(env)

    env.wm_title("Layrageu")
    env.geometry(starting_dimensions())

    # --- Running the GUI ------------------------------------------------

    env.mainloop()
    sys.exit(0)


# Programme =============================================================#

if __name__ == "__main__":
    main()

# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# +--------------------------------------------------------------------------+
# | Layrageu                                                                 |
# | Copyright (C) 2022 Étienne Nadji                                         |
# |                                                                          |
# | This program is free software: you can redistribute it and/or modify     |
# | it under the terms of the GNU Affero General Public License as           |
# | published by the Free Software Foundation, either version 3 of the       |
# | License, or (at your option) any later version.                          |
# |                                                                          |
# | This program is distributed in the hope that it will be useful,          |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of           |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            |
# | GNU Affero General Public License for more details.                      |
# |                                                                          |
# | You should have received a copy of the GNU Affero General Public License |
# | along with this program.  If not, see <https://www.gnu.org/licenses/>.   |
# +--------------------------------------------------------------------------+

"""
Layrageu : settings.
"""

# Imports ===============================================================#

# Standard library ------------------------------

import configparser

from pathlib import Path
from typing import Literal, NoReturn, List
from collections import OrderedDict

# Third parties ---------------------------------

from slugify import slugify

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

TootVisibility = Literal["private", "unlisted", "public"]

# Functions =============================================================#


def slugify_instance_url(url: str) -> str:
    """
    Slugify an instance URL. Used as a better default name for a new
    instance.

    `https://mastodon.social` ⇒ `mastodon-social`

    :type url: str
    :param url: Instance URL
    :rtype: str
    """

    unprotocoled = url.split("//")[1:]
    slug_base = "/".join(unprotocoled)

    try:
        slug = slugify(slug_base)
    except NameError:
        # Fallback to slugify use of unicode()
        slug = slug_base.replace("/", "-")
        slug = slug.replace(".", "-")

    return slug


def _bool_to_string(value: bool) -> str:
    """
    Converts boolean to "yes" or "no" string.

    :type value: bool
    :param value: Boolean to convert.
    :rtype: bool
    :returns: "yes" or "no" string.
    """
    return {True: "yes", False: "no"}[bool(value)]


def _string_to_bool(string: str, default: bool = False) -> bool:
    """
    Converts a "yes" or "no" string to a boolean value.

    :type string: str
    :param string: "yes" or "no" string.
    :rtype: bool
    """

    try:
        value = {"yes": True, "no": False}[str(string).lower()]
    except IndexError:
        value = default

    return value


# Classes ===============================================================#


class UnknownInstance(Exception):
    """
    Exception raised when a Mastodon instance is not known by Layrageu's
    settings.
    """


class MastodonInstance:
    """
    Mastodon instance class.
    """

    def __init__(
        self,
        name: str,
        url: str,
        usermail: str,
        first_tootv: TootVisibility = "private",
        other_tootv: TootVisibility = "private",
        always_sensitive: bool = False,
    ):
        self.name = name
        self.url = url
        self.usermail = usermail

        self.first_tootv = first_tootv
        self.other_tootv = other_tootv

        self.always_sensitive = always_sensitive

    def slugifyed_url(self) -> str:
        """
        Return a slug of this instance URL.

        :rtype: str
        """
        return slugify_instance_url(self.url)

    def as_dict(self) -> dict:
        """
        ReturnBoolean to convert.s self as a dictionnary.

        :rtype: bool
        """

        return {
            "url": self.url,
            "user": self.usermail,
            "tootv": self.other_tootv,
            "init_tootv": self.first_tootv,
            "always_sensitive": _bool_to_string(self.always_sensitive),
        }

    def __repr__(self):
        return str(self.as_dict())

    def __eq__(self, other) -> bool:
        if not isinstance(other, MastodonInstance):
            raise TypeError(
                "Not comparing self with a MastodonInstance instance."
            )

        return self.__repr__() == other.__repr__()

    def update_from_cfg(self, instance_data: dict) -> NoReturn:
        """
        :type instance_data: dict
        :param instance_data: Dictionnary of datas from ConfigParser.
        """

        def tootv_or_default(value: str, default: str) -> str:
            if value in ["private", "unlisted", "public"]:
                return value

            return default

        for option, value in instance_data.items():
            if option == "url":
                self.url = value
            if option == "user":
                self.usermail = value
            if option == "init_tootv":
                self.first_tootv = tootv_or_default(value, "private")
            if option == "tootv":
                self.other_tootv = tootv_or_default(value, "private")
            if option == "always_sensitive":
                self.always_sensitive = _string_to_bool(value)


class Settings:
    """
    Layrageu's settings file class.

    :type folder: pathlib.Path
    :param folder: Settings folder
    """

    def __init__(self, folder: Path):
        self.apps = OrderedDict()
        self.instances = OrderedDict()

        # Create the settings folder if it doesn't exists ----------------

        self.folder = Path(folder).resolve(False)
        self.folder.mkdir(exist_ok=True)

        # Transition from legacy settings file ---------------------------

        legacy = self.folder / "settings.ini"

        if legacy.exists():
            legacy.rename(self.folder / "instances.ini")

        # ----------------------------------------------------------------

        self.files = {
            "apps": self.folder / "apps.ini",
            "instances": self.folder / "instances.ini",
        }

        as_new = not self.files["instances"].exists()

        # ----------------------------------------------------------------

        if as_new:
            self.create()
        else:
            self.load()

    # =======================================================================

    def instances_count(self) -> int:
        """
        Returns count of known instances.

        :rtype: int
        """
        return len(self.instances)

    def instances_names(self) -> List[str]:
        """
        Returns names of known instances.

        :rtype: List[str]
        """
        return list(self.instances.keys())

    # =======================================================================

    def needs_first_user_action(self) -> bool:
        """
        Returns True if there is only one access configured, which is the
        default (and unusable) access.

        :rtype: bool
        """

        if self.instances_count() > 1:
            return False

        default = MastodonInstance(
            "default",
            "https://example.com",
            "user@mail.example.com",
            "public",
            "unlisted",
            False,
        )

        return [v for v in self.instances.values()][0] == default

    # Save / Load / Create from scratch the settings ========================

    def save(self) -> bool:
        """
        Save the settings.

        :rtype: bool
        """

        # Instances settings ---------------------------------------------

        parser = configparser.ConfigParser()

        for name, instance in self.instances.items():
            for option_name, option_value in instance.as_dict().items():
                if option_name == "name":
                    continue

                if name not in parser:
                    parser[name] = {}

                parser[name][option_name] = option_value

        with open(self.files["instances"], "w", encoding="utf8") as stfile:
            parser.write(stfile)

        # ----------------------------------------------------------------

        return True

    def load(self) -> bool:
        """
        Load settings.

        :rtype: bool
        """

        # Instances settings ---------------------------------------------

        parser = configparser.ConfigParser()
        parser.read(self.files["instances"])

        for instance in parser.sections():
            new_instance = MastodonInstance(
                instance,
                None,
                None,
                "private",
                "private",
                False,
            )

            new_instance.update_from_cfg(parser[instance])
            self.instances[instance] = new_instance

        # ----------------------------------------------------------------

        return True

    def create(self) -> NoReturn:
        """
        Creates a new settings folder with default datas.
        """

        self.__add_default_instance()
        self.save()

    # Mastodon instances management =========================================

    def add_instance(
        self,
        name: str,
        url: str,
        usermail: str,
        first_tootv: TootVisibility = "private",
        other_tootv: TootVisibility = "private",
        always_sensitive: bool = False,
    ) -> NoReturn:
        """
        Add a new Mastodon instance access.

        :type name: str
        :param name: Mastodon instance name.
        :type url: str
        :param url: Mastodon instance URL.
        :type usermail: str
        :param usermail: Mastodon instance user mail.
        :type first_tootv: Literal["private", "unlisted", "public"]
        :param first_tootv: Visibility of the first toot.
        :type other_tootv: Literal["private", "unlisted", "public"]
        :param other_tootv: Visibility of non-first toots.
        :type always_sensitive: bool
        :param always_sensitive: Always mark toots as sensitive.
        """

        new_instance = MastodonInstance(
            name,
            url,
            usermail,
            first_tootv,
            other_tootv,
            always_sensitive,
        )

        self.instances[name] = new_instance

    def __add_default_instance(self) -> NoReturn:
        """
        Add a default and unusable Mastodon instance access.
        """

        self.add_instance(
            name="default",
            url="https://example.com",
            usermail="user@mail.example.com",
            first_tootv="public",
            other_tootv="unlisted",
            always_sensitive=False,
        )

    def mi_delete(self, instance_name: str) -> NoReturn:
        """
        Delete a Mastodon instance access.
        """

        self.instances.pop(instance_name)

        if not self.instances:
            self.__add_default_instance()

        self.save()

    def mi_exists(self, instance_name: str) -> bool:
        """
        Does settings has an instance named `instance_name` ?

        :type instance_name: str
        :param instance_name: Name of the instance to check.
        :rtype: bool
        """

        return self.instances.get(instance_name) is not None

    def mi_url(self, instance_name: str) -> str:
        """
        Returns the URL of the instance called `instance_name`

        :type instance_name: str
        :param instance_name: Name of the instance.
        :rtype: str
        :returns: Instance's URL
        """

        if self.mi_exists(instance_name):
            return self.instances[instance_name].url

        raise UnknownInstance()

    def mi_user(self, instance_name: str) -> str:
        """
        Returns the user mail of the instance called `instance_name`

        :type instance_name: str
        :param instance_name: Name of the instance.
        :rtype: str
        :returns: Instance's user mail
        """

        if self.mi_exists(instance_name):
            return self.instances[instance_name].usermail

        raise UnknownInstance()

    def mi_init_tootv(self, instance_name: str) -> str:
        """
        Returns the visibility of the first toot of a thread posted on the
        instance called `instance_name`

        :type instance_name: str
        :param instance_name: Name of the instance.
        :rtype: str
        :returns: First toot visibility
        """

        if self.mi_exists(instance_name):
            return self.instances[instance_name].first_tootv

        raise UnknownInstance()

    def mi_tootv(self, instance_name: str) -> str:
        """
        Returns the visibility of the non-first toots of a thread posted on
        the instance called `instance_name`

        :type instance_name: str
        :param instance_name: Name of the instance.
        :rtype: str
        :returns: Non-first toot visibility
        """

        if self.mi_exists(instance_name):
            return self.instances[instance_name].other_tootv

        raise UnknownInstance()

    def mi_always_sensitive(self, instance_name: str) -> bool:
        """
        Returns whether toots should be marked as sensitive on the instance
        called `instance_name`.

        :type instance_name: str
        :param instance_name: Name of the instance.
        :rtype: bool
        :returns: Always mark toots as sensitive?
        """

        if self.mi_exists(instance_name):
            return self.instances[instance_name].always_sensitive

        raise UnknownInstance()


# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

"""
Layrageu : XSL stylesheet as a module file.
"""

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

FODT_TO_TEXT_XML = """<xsl:stylesheet version="1.0"
  xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
  xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0"
  xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
  xmlns:xlink="http://www.w3.org/1999/xlink">

<xsl:output omit-xml-declaration="yes" method="text" encoding="UTF-8" />

<xsl:strip-space elements="*" />

<xsl:variable name="emph-style">Emphasis</xsl:variable>
<xsl:variable name="bold-style">Strong_20_Emphasis</xsl:variable>

<xsl:template match="/"><xsl:apply-templates select="/office:document/office:body/office:text" /></xsl:template>

<xsl:template match="text:a"><xsl:apply-templates /></xsl:template>
<xsl:template match="text:line-break">{lb}</xsl:template>

<xsl:template match="text:p"><xsl:apply-templates /><xsl:text>
</xsl:text></xsl:template>

<xsl:template match="text:span">
  <xsl:variable name="format">
    <xsl:choose>
      <xsl:when test="@text:style-name = $bold-style">bold</xsl:when>
      <xsl:when test="@text:style-name = $emph-style">emph</xsl:when>
      <xsl:otherwise>unknown</xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <xsl:choose>
    <xsl:when test="$format = 'bold'">**</xsl:when>
    <xsl:when test="$format = 'emph'">*</xsl:when>
    <xsl:otherwise></xsl:otherwise>
  </xsl:choose>
  <xsl:apply-templates />
  <xsl:choose>
    <xsl:when test="$format = 'bold'">**</xsl:when>
    <xsl:when test="$format = 'emph'">*</xsl:when>
    <xsl:otherwise></xsl:otherwise>
  </xsl:choose>
</xsl:template>

</xsl:stylesheet>"""

# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

VERSION = "0.2.1"
DESCRIPTION = "Send Mastodon threads from Flat OpenDocument files."

REQUIRED = [
    "lxml",
    "rich",
    "docopt",
    "appdirs",
    "slugify",
    "iso639-lang",
    "Mastodon.py",
    "ttkSimpleDialog",
]

setuptools.setup(
    name="layrageu",
    version=VERSION,
    author="Étienne Nadji",
    author_email="etnadji@eml.cc",
    description=DESCRIPTION,
    long_description=long_description,
    packages=setuptools.find_packages(),
    platforms="any",
    license="GNU Affero General Public License v3 or later (AGPLv3+)",
    classifiers=[
        "Topic :: Communications",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Markup :: XML",
        "Intended Audience :: Education",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
    ],
    project_urls={
        "Source Code": "https://framagit.org/etnadji/layrageu",
        "Issue tracker": "https://framagit.org/etnadji/layrageu/-/issues",
        "Documentation": "https://etnadji.fr/rsc/layrageu/doc",
    },
    entry_points={
        "console_scripts": [
            "layrageu=layrageu.tui:main",
            "layrageu-gui=layrageu.gui:main",
        ]
    },
    python_requires=">=3.8",
    install_requires=REQUIRED,
    keywords=["Mastodon", "LibreOffice", "OpenDocument"],
)

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# +--------------------------------------------------------------------------+
# | Layrageu                                                                 |
# | Copyright (C) 2022 Étienne Nadji                                         |
# |                                                                          |
# | This program is free software: you can redistribute it and/or modify     |
# | it under the terms of the GNU Affero General Public License as           |
# | published by the Free Software Foundation, either version 3 of the       |
# | License, or (at your option) any later version.                          |
# |                                                                          |
# | This program is distributed in the hope that it will be useful,          |
# | but WITHOUT ANY WARRANTY; without even the implied warranty of           |
# | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            |
# | GNU Affero General Public License for more details.                      |
# |                                                                          |
# | You should have received a copy of the GNU Affero General Public License |
# | along with this program.  If not, see <https://www.gnu.org/licenses/>.   |
# +--------------------------------------------------------------------------+

"""Layrageu : send Mastodon toots from Flat OpenDocument files.

Usage:
  tui.py send <document> [--instance=name] [--cw=title] [--sensitive]
                         [--preview] [--nopager] [--verbose]
  tui.py preview <document> [--nopager] [--verbose]
  tui.py configure [--name=<name>] [--url=<url>] [--user=<mail>]
                   [--ftv=<toot_visibility>] [--ntv=<toots_visibility>]
                   [--always-sensitive] [--verbose]
  tui.py version

Options:
  -h --help                Displays help.
  --verbose                Verbosity of the command.
  --preview                Preview toots before sending.
  --nopager                Always disable pager on toots preview.
  --sensitive              Marks the thread's toots as sensitive.
  --url=<url>              URL of the instance.
  --user=<mail>            User mail for the instance.
  --name=<name>            Name of the instance. [default: default]
  --ftv=<toot_visibility>  Visibility of first toot. Must be private,
                           unlisted or public. public by default.
  --ntv=<toots_visibility> Visibility of following toots. Must be private,
                           unlisted or public. unlisted by default.
  --always-sensitive       Posts on this instance will be marked as sensitive.

Layrageu. Copyright (C) 2022 Étienne Nadji.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.
"""

# Imports ===============================================================#

# Standard library ------------------------------

import os
import sys
import getpass

from pathlib import Path
from typing import List

# Third parties ---------------------------------

import docopt

from appdirs import user_config_dir

from rich.panel import Panel
from rich.console import Console
from rich.prompt import Prompt, Confirm

# Layrageu --------------------------------------

from layrageu import TOOT_VISIBILITIES, debug_mode
from layrageu.loader import convert, Toot
from layrageu.sender import send_toots
from layrageu.settings import Settings

# Variables globales ====================================================#

__author__ = "Étienne Nadji <etnadji@eml.cc>"
__version__ = "0.2.1"

# Classes ===============================================================#


class Layrageu:
    COMMANDS = ["preview", "send", "configure", "version"]

    def __init__(
        self,
        arguments: dict,
    ):
        # CLI arguments and debug mode -----------------------------------

        self.arguments = arguments
        self.debug_mode = debug_mode()

        # Load Layrageu's settings ---------------------------------------

        self.settings = Settings(user_config_dir("layrageu"))

        # Find which command is used -------------------------------------

        self.command = False

        for command in Layrageu.COMMANDS:
            if self.arguments[command]:
                self.command = command
                break

        # Initializes rich Console ---------------------------------------

        self.console = Console()

    def _load_document(self) -> List[Toot]:
        """
        Load FODT documents as Mastodon toots.
        """
        return convert(
            self.arguments["<document>"], self.arguments["--verbose"]
        )

    def wizard(self) -> bool:
        """
        Instance access configuration wizard for the first time the user
        use a command that requires instance access.

        :rtype: bool
        :returns: Configuration success.
        """

        self.console.print(
            Panel("[yellow]Layrageu configuration wizard[/yellow]", width=33)
        )
        self.console.print(
            "\nYou didn't configure your Mastodon instance access.\n"
        )

        do_config = Confirm.ask("Would you like to configure it now?")

        if not do_config:
            return False

        clear = False
        wizard_loop = True

        url = None
        usermail = None
        first_tootv = "private"
        other_tootv = "private"

        while wizard_loop:

            if clear:
                self.console.clear()
            else:
                print()

            url = Prompt.ask("Instance URL")

            if not url:
                self.console.print("[red]Empty instance URL[/red]")
                continue

            usermail = Prompt.ask("User mail")

            if not usermail:
                self.console.print("[red]Empty user mail[/red]")
                continue

            first_tootv = Prompt.ask(
                "First toot visibility",
                choices=TOOT_VISIBILITIES,
                default="private",
            )
            other_tootv = Prompt.ask(
                "Next toots visibility",
                choices=TOOT_VISIBILITIES,
                default="unlisted",
            )

            summary_width = len(url) + 20

            print()

            self.console.print(
                Panel(
                    "\n".join(
                        [
                            "[blue]Access[/blue]",
                            "",
                            f"[yellow]User mail[/yellow] : {usermail}",
                            f"[yellow]Instance URL[/yellow] : {url}",
                            "",
                            "[blue]Toots visibilities[/blue]",
                            "",
                            f"[yellow]First toot[/yellow] : {first_tootv}",
                            f"[yellow]Next toots[/yellow] : {other_tootv}",
                        ]
                    ),
                    width=summary_width,
                )
            )

            print()

            do_set = Confirm.ask("Are you OK with those options ?")

            if do_set:
                wizard_loop = False
            else:
                clear = True

        self.console.clear()

        url = None
        usermail = None
        first_tootv = "private"
        other_tootv = "private"

        instance = self.settings.instances[-1]

        instance.url = url
        instance.usermail = usermail
        instance.first_tootv = first_tootv
        instance.other_tootv = other_tootv

        return True

    def run(self) -> int:
        methods = {
            "send": self.cmd_send,
            "preview": self.cmd_preview,
            "version": self.cmd_version,
            "configure": self.cmd_configure,
        }

        return methods[self.command]()

    def cmd_version(self) -> int:
        print(__version__)
        return 0

    def cmd_preview(self) -> int:
        """
        preview command : Preview toots.
        """

        def toot_as_preview(toot_index, toot):
            return Panel(
                toot.as_markdown(),
                width=80,
                title=f"\nToot {toot_index+1} ({len(toot)})",
            )

        toots = self._load_document()

        if self.arguments["--nopager"]:
            use_pager = False
        else:
            use_pager = sum([len(toot) for toot in toots]) >= 1500

        if use_pager:
            print("[Previewed in terminal pager]")

            with self.console.pager():
                for idx, toot in enumerate(toots):
                    self.console.print(toot_as_preview(idx, toot))

            return 0

        for idx, toot in enumerate(toots):
            self.console.print(toot_as_preview(idx, toot))

        return 0

    def cmd_send(self) -> int:
        """
        send command : Send toots to Mastodon instance.
        """

        toots = self._load_document()

        # Preview the toots before sending them --------------------------

        if self.arguments["--preview"]:
            self.cmd_preview()
            print()

        # First-time configuration wizard --------------------------------

        if self.settings.needs_first_user_action():
            configured = self.wizard()

            if not configured:
                return 1

        # Check selected instance ----------------------------------------

        instance_name = "default"

        if self.arguments["--instance"] is not None:
            instance_name = self.arguments["--instance"]

        if not self.settings.mi_exists(instance_name):
            print(
                f"Critical: Mastodon instance “{self.arguments['--instance']}” "
                "does not exists in Layrageu's configuration."
            )
            return 1

        # ----------------------------------------------------------------

        try:
            password = self.console.input(
                "[red]Password[red] : ", password=True
            )

            result = send_toots(
                instance_name,
                self.settings,
                toots,
                password,
                self.arguments["--cw"],
                self.arguments["--sensitive"],
                terminal=True,
            )
        except KeyboardInterrupt:
            return 0

        # ----------------------------------------------------------------

        return {True: 0, False: 1}[result]

    def cmd_configure(self) -> int:
        """
        configure command : create or modify a Mastodon instance.
        """

        opt_errors = False
        has_instance = self.settings.mi_exists(self.arguments["--name"])

        # Default value is not defined with docopt string, because configure
        # command is used to create AND modify instances. So if toot visibility
        # options were set, the user would override those options by mistake.
        if not has_instance:
            if self.arguments["--ftv"] is None:
                self.arguments["--ftv"] = "public"
            if self.arguments["--ntv"] is None:
                self.arguments["--ntv"] = "unlisted"

        # Check if visibility options use the correct set of values.
        for option in ["--ftv", "--ntv"]:
            if self.arguments[option] in TOOT_VISIBILITIES:
                continue

            print(
                f"Wrong option value : {option} must be "
                "private, unlisted or public (without quotes).\n\t"
                f"Got '{self.arguments[option]}'."
            )
            opt_errors = True

        # Check if mandatory options when creating an instance are not empty.
        if not has_instance:
            for option in ["--url", "--user"]:
                if self.arguments[option] is not None:
                    continue

                print(
                    f"Wrong option value : {option} is mandatory, yet empty."
                )
                opt_errors = True

        # Options are so wrong that Layrageu don't want to live anymore.
        if opt_errors:
            return 1

        # Modify a Mastodon instance -------------------------------------

        if has_instance:
            instance_name = self.arguments["--name"]

            if self.arguments["--url"] is not None:
                self.settings.instances[instance_name].url = self.arguments[
                    "--url"
                ]

            if self.arguments["--user"] is not None:
                self.settings.instances[
                    instance_name
                ].usermail = self.arguments["--user"]

            if self.arguments["--ftv"]:
                self.settings.instances[
                    instance_name
                ].first_tootv = self.arguments["--ftv"]

            if self.arguments["--ntv"]:
                self.settings.instances[
                    instance_name
                ].other_tootv = self.arguments["--ntv"]

            self.settings.instances[
                instance_name
            ].always_sensitive = self.arguments["--always-sensitive"]

            self.settings.save()

            return 0

        # Creating a Mastodon instance -----------------------------------

        self.settings.add_instance(
            self.arguments["--name"],
            self.arguments["--url"],
            self.arguments["--user"],
            self.arguments["--ftv"],
            self.arguments["--ntv"],
            self.arguments["--always-sensitive"],
        )

        self.settings.save()

        # ----------------------------------------------------------------

        return 0


# Fonctions =============================================================#


def main():
    """
    Layrageu command line entry point.
    """

    exitcode = 0

    ARGUMENTS = docopt.docopt(__doc__, version=__version__)

    if debug_mode():
        print("\nCLI arguments :", ARGUMENTS, "\n")

    app = Layrageu(ARGUMENTS)

    exitcode = app.run()

    sys.exit(exitcode)


# Programme =============================================================#

if __name__ == "__main__":
    main()


# vim:set shiftwidth=4 softtabstop=4:
